/* 
   $Id: sgwindow.c,v 1.5 2001/06/27 02:54:32 sho Exp $

   Copyright (C) 2001 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of sgwm.

   sgwm is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   sgwm is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with sgwm; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "sgwm.h"
#include <gdk/gdkx.h>
#include <X11/Xatom.h>
#include "sgwindow.h"
#include "sgwmwidget.h"
#include "atoms.h"
#include "sgbarwindow.h"

static void sgwindow_class_init(SGWindowClass *klass);
static void sgwindow_init(SGWindow *win);

GtkType sgwindow_get_type(void)
{
    static GtkType sgwindow_type = 0;

    if (sgwindow_type == 0) {
	GtkTypeInfo sgwindow_info = {
	    "SGWindow",
	    sizeof(SGWindow),
	    sizeof(SGWindowClass),
	    (GtkClassInitFunc)sgwindow_class_init,
	    (GtkObjectInitFunc)sgwindow_init,
	    NULL,
	    NULL,
	    NULL
	};
	sgwindow_type = gtk_type_unique(gtk_widget_get_type(),
	    &sgwindow_info);
    }

    return sgwindow_type;
}

void sgwindow_destroy_notify(SGWindow *sgwin)
{
    GtkWidget *widget = GTK_WIDGET(sgwin);

    GTK_WIDGET_UNSET_FLAGS(widget,GTK_REALIZED);
    GTK_WIDGET_UNSET_FLAGS(widget,GTK_MAPPED);
    GTK_WIDGET_UNSET_FLAGS(widget,GTK_VISIBLE);
    if (widget->window == NULL) {
	return;
    }
    if (sgwin->type == SGWINDOW_APP) {
	Atom rt = None;
	int rf;
	unsigned long rni, bar;
	unsigned char *rv;
	Window xwin;
	GdkWindow *win;

	gdk_error_trap_push();
	rt = 0;
	xwin = GDK_WINDOW_XWINDOW(widget->window);
	XGetWindowProperty(GDK_DISPLAY(),xwin,xa_sg_menubar,0,1,
	  FALSE,XA_WINDOW,&rt,&rf,&rni,&bar,&rv);
	if (rt == XA_WINDOW) {
	    if (rf == 32 && rni == 1) {
		Window rxwin = ((Window *)rv)[0];
		if ((win = gdk_window_lookup(rxwin)) != NULL) {
		    GtkWidget *tw;
		    gdk_window_get_user_data(win,(gpointer *)&tw);
		    sgwindow_destroy_notify(SGWINDOW(tw));
		    gtk_widget_unref(tw);
		}
	    }
	    XFree(rv);
	}
	rt = 0;
	XGetWindowProperty(GDK_DISPLAY(),xwin,xa_sg_toolbar,
	  0,1,FALSE,XA_WINDOW,&rt,&rf,&rni,&bar,&rv);
	if (rt == XA_WINDOW && rf == 32 && rni == 1) {
	    if (rf == 32 && rni == 1) {
		Window rxwin = ((Window *)rv)[0];
		if ((win = gdk_window_lookup(rxwin)) != NULL) {
		    GtkWidget *tw;
		    gdk_window_get_user_data(win,(gpointer *)&tw);
		    sgwindow_destroy_notify(SGWINDOW(tw));
		    gtk_widget_unref(tw);
		}
	    }
	    XFree(rv);
	}
	gdk_flush();
	if (gdk_error_trap_pop()) {
	    ;
	}
    } else {
	gdk_error_trap_push();
	XRemoveFromSaveSet(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(widget->window));
	gdk_flush();
	if (gdk_error_trap_pop()) {
	    ;
	}
    }
    gdk_window_destroy_notify(widget->window);
    widget->window = NULL;
}

static void sgwindow_realize(GtkWidget *widget)
{
    if (widget->window == 0) return;
    GTK_WIDGET_SET_FLAGS(widget,GTK_REALIZED);
}

static void sgwindow_unrealize(GtkWidget *widget)
{
    if (!GTK_WIDGET_REALIZED(widget)) return;
    GTK_WIDGET_UNSET_FLAGS(widget,GTK_REALIZED);
}

static void sgwindow_map(GtkWidget *widget)
{
    SGWindow *sgwin;

    if (widget->window == 0) return;
    GTK_WIDGET_SET_FLAGS(widget,GTK_MAPPED);
    sgwin = SGWINDOW(widget);
    gdk_error_trap_push();
    gdk_window_show(widget->window);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(SGWINDOW(widget));
    }
}

static void sgwindow_unmap(GtkWidget *widget)
{
    SGWindow *sgwin;

    if (widget->window == 0) return;
    GTK_WIDGET_UNSET_FLAGS(widget,GTK_MAPPED);
    sgwin = SGWINDOW(widget);
    gdk_error_trap_push();
    gdk_window_hide(widget->window);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(SGWINDOW(widget));
    }
}

static void sgwindow_size_request(GtkWidget *widget, GtkRequisition *req)
{
    req->width = widget->requisition.width;
    req->height = widget->requisition.height;
}

void sgwindow_raise(SGWindow *sgwin)
{
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return;
    gdk_error_trap_push();
    gdk_window_raise(widget->window);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(SGWINDOW(widget));
    }
}

void sgwindow_move(SGWindow *sgwin, gint x, gint y)
{
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return;
    gdk_error_trap_push();
    gdk_window_move(widget->window,x,y);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(SGWINDOW(widget));
    }
    widget->allocation.x = x;
    widget->allocation.y = y;
}

void sgwindow_move_resize(SGWindow *sgwin, gint x, gint y, gint w, gint h)
{
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return;
    gdk_error_trap_push();
    gdk_window_move_resize(widget->window,x,y,w,h);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(SGWINDOW(widget));
    }
    widget->allocation.x = x;
    widget->allocation.y = y;
    widget->requisition.width = widget->allocation.width = w;
    widget->requisition.height = widget->allocation.height = h;
}

void sgwindow_resize(SGWindow *sgwin, gint w, gint h)
{
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return;
    gdk_error_trap_push();
    gdk_window_resize(widget->window,w,h);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(SGWINDOW(widget));
    }
    widget->requisition.width = widget->allocation.width = w;
    widget->requisition.height = widget->allocation.height = h;
}

static void sgwindow_parent_set(GtkWidget *widget, GtkWidget *pre_parent)
{
    if (pre_parent != widget->parent && widget->parent != NULL) {
	gint x,y;
	GdkWindow *pwin;

	if (widget->window == NULL) return;
	gdk_error_trap_push();
	gdk_window_get_position(widget->window,&x,&y);
	pwin = gtk_widget_get_parent_window(widget);
	gdk_window_reparent(widget->window, pwin,x,y);
	gdk_flush();
	if (gdk_error_trap_pop()) {
	    sgwindow_destroy_notify(SGWINDOW(widget));
	}
    }
}

static void sgwindow_size_allocate(GtkWidget *widget, GtkAllocation *alloc)
{
    SGWindow *sgwin;
    widget->allocation = *alloc;

    if (widget->window == NULL) return;
    sgwin = SGWINDOW(widget);
    gdk_error_trap_push();
    sgwindow_move_resize(SGWINDOW(widget),alloc->x,alloc->y,
      alloc->width,alloc->height);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(SGWINDOW(widget));
    }
}

static void sgwindow_class_init(SGWindowClass *klass)
{
    GtkWidgetClass *widget_class;
    GtkObjectClass *object_class;

    object_class = (GtkObjectClass *)klass;

    widget_class = (GtkWidgetClass *)klass;
    widget_class->realize = sgwindow_realize;
    widget_class->size_allocate = sgwindow_size_allocate;
    widget_class->unrealize = sgwindow_unrealize;
    widget_class->map = sgwindow_map;
    widget_class->unmap = sgwindow_unmap;
    widget_class->size_request = sgwindow_size_request;
    widget_class->parent_set = sgwindow_parent_set;
}

static void sgwindow_init(SGWindow *win)
{
    win->workspace = -1;
    win->type = SGWINDOW_APP;
}

GtkWidget *sgwindow_setup(SGWindow *sgwin, guint32 xid, SGWMWidget *sgwm)
{
    GtkWidget *widget;
    Window rw;
    gint x,y,w,h,bw;
    guint d;

    widget  = (GtkWidget *)sgwin;
    sgwin->sgwm = sgwm;
    gdk_error_trap_push();
    XGetGeometry(GDK_DISPLAY(),xid,&rw,&x,&y,&w,&h,&bw,&d);
    sgwin->border_width = bw;
    widget->allocation.x = x;
    widget->allocation.y = y;
    widget->requisition.width = widget->allocation.width = w;
    widget->requisition.height = widget->allocation.height = h;
    widget->window = gdk_window_foreign_new(xid);
    if (widget->window != NULL) gdk_window_set_user_data(widget->window,widget);
    GTK_WIDGET_SET_FLAGS(widget,GTK_REALIZED);
    gdk_flush();
    if (gdk_error_trap_pop() || widget->window == NULL) {
	sgwindow_destroy_notify(SGWINDOW(widget));
	gtk_widget_destroy(widget);
	return NULL;
    }
    return widget;
}

GtkWidget *sgwindow_new(guint32 xid, SGWMWidget *sgwm)
{
    GtkWidget *widget;
    SGWindow *sgwin;

    widget  = GTK_WIDGET(gtk_type_new(sgwindow_get_type()));
    sgwin = (SGWindow *)widget;
    return sgwindow_setup(sgwin,xid,sgwm);
}

GtkWidget *sgwindow_from_gdk_window(GdkWindow *win, guint32 xid, SGWMWidget *sgwm)
{
    GtkWidget *widget;

    gdk_error_trap_push();
    if (win == NULL) {
	widget = sgwindow_new(xid,sgwm);
    } else {
	gdk_window_get_user_data(win,(gpointer *)&widget);
	if (IS_SGBARWINDOW(widget)) {
	    widget = NULL;
	}
    }
    gdk_flush();
    if (gdk_error_trap_pop()) {
	return NULL;
    }
    return widget;
}

GtkWidget *sgwindow_lookup(guint32 xid, SGWMWidget *sgwm)
{
    GdkWindow *win;

    win = gdk_window_lookup(xid);
    return sgwindow_from_gdk_window(win,xid,sgwm);
}

void sgwindow_get_bars(SGWindow *sgwin, GtkWidget **menubar, GtkWidget **toolbar)
{
    Atom rt = None;
    int rf;
    unsigned long rni, bar;
    unsigned char *rv;
    GtkWidget *widget = GTK_WIDGET(sgwin);
    Window xwin;

    *menubar = *toolbar = NULL;
    if (widget->window == NULL) return;
    gdk_error_trap_push();
    rt = 0;
    xwin = GDK_WINDOW_XWINDOW(widget->window);
    XGetWindowProperty(GDK_DISPLAY(),xwin,xa_sg_menubar,0,1,
      FALSE,XA_WINDOW,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_WINDOW) {
	if (rf == 32 && rni == 1) {
	    Window rxwin = ((Window *)rv)[0];
	    *menubar = sgbarwindow_lookup(rxwin,sgwin->sgwm,SGWINDOW_MENUBAR);
	    if (*menubar != NULL) {
		gtk_widget_ref(*menubar);
	    }
	}
	XFree(rv);
    }
    rt = 0;
    XGetWindowProperty(GDK_DISPLAY(),xwin,xa_sg_toolbar,
      0,1,FALSE,XA_WINDOW,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_WINDOW && rf == 32 && rni == 1) {
	if (rf == 32 && rni == 1) {
	    Window rxwin = ((Window *)rv)[0];
	    *toolbar = sgbarwindow_lookup(rxwin,sgwin->sgwm,SGWINDOW_TOOLBAR);
	    if (*toolbar != NULL) {
		gtk_widget_ref(*toolbar);
	    }
	}
	XFree(rv);
    }
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(sgwin);
    }
}

void sgwindow_change_win_state(SGWindow *sgwin, gint state)
{
    unsigned int data[1];
    GtkWidget *widget = GTK_WIDGET(sgwin);

    data[0] = (unsigned int)state;
    if (widget->window == NULL) return;
    gdk_error_trap_push();
    XChangeProperty(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(widget->window),xa_win_state,
      XA_CARDINAL,32,PropModeReplace,(unsigned char *)data,1);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(sgwin);
    }
}

void sgwindow_change_wm_state(SGWindow *sgwin, int state)
{
    unsigned int data[2];
    GtkWidget *widget = GTK_WIDGET(sgwin);

    data[0] = (unsigned int)state;
    data[1] = 0;
    if (widget->window == NULL) return;
    gdk_error_trap_push();
    XChangeProperty(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(widget->window),xa_wm_state,
      xa_wm_state,32,PropModeReplace,(unsigned char *)data,2);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(sgwin);
    }
}

gboolean sgwindow_is_same_leader(SGWindow *sgwin1, SGWindow *sgwin2)
{
    Window l1 = 0;
    Window l2 = 0;
    Atom rt = None;
    int rf;
    unsigned long rni, bar;
    unsigned char *rv;
    GtkWidget *widget1 = GTK_WIDGET(sgwin1);
    GtkWidget *widget2 = GTK_WIDGET(sgwin2);

    if (widget1->window == NULL
      || widget2->window == NULL) return FALSE;
    gdk_error_trap_push();
    rt = 0;
    XGetWindowProperty(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(widget1->window),
      xa_wm_client_leader,0,1,
      FALSE,XA_WINDOW,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_WINDOW) {
	if (rf == 32 && rni == 1) {
	    l1 = ((Window *)rv)[0];
	}
	XFree(rv);
    }
    rt = 0;
    XGetWindowProperty(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(widget2->window),
      xa_wm_client_leader,0,1,
      FALSE,XA_WINDOW,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_WINDOW) {
	if (rf == 32 && rni == 1) {
	    l2 = ((Window *)rv)[0];
	}
	XFree(rv);
    }
    if (gdk_error_trap_pop()) {
	return FALSE;
    }
    if (l1 == 0 || l2 == 0) return FALSE;
    return l1 == l2;
}

Window sgwindow_get_transient_for(SGWindow *sgwin)
{
    Window rwin = 0;
    Atom rt = None;
    int rf;
    unsigned long rni, bar;
    unsigned char *rv;
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return 0;
    gdk_error_trap_push();
    rt = 0;
    XGetWindowProperty(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(widget->window),
      XA_WM_TRANSIENT_FOR,0,1,
      FALSE,XA_WINDOW,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_WINDOW) {
	if (rf == 32 && rni == 1) {
	    rwin = ((Window *)rv)[0];
	}
	XFree(rv);
    }
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(sgwin);
    }
    return rwin;
}

void sgwindow_set_focus(SGWindow *sgwin)
{
    Atom *protocols = NULL;
    int i,n;
    GtkWidget *widget;
    Window xwin;

    if (sgwin == NULL) {
	XSetInputFocus(GDK_DISPLAY(),PointerRoot,RevertToPointerRoot, CurrentTime);
	return;
    }
    widget = GTK_WIDGET(sgwin);
    if (widget->window == NULL) return;
    xwin = GDK_WINDOW_XWINDOW(widget->window);
    gdk_error_trap_push();
    XSetInputFocus(GDK_DISPLAY(),xwin,RevertToPointerRoot, CurrentTime);
    if (XGetWMProtocols(GDK_DISPLAY(),xwin,&protocols,&n)) {
	for (i = 0;i < n;i++) {
	    if (protocols[i] == xa_wm_take_focus) {
		XClientMessageEvent ev;

		ev.type = ClientMessage;
		ev.window = xwin;
		ev.message_type = xa_wm_protocols;
		ev.format = 32;
		ev.data.l[0] = xa_wm_take_focus;
		ev.data.l[1] = CurrentTime;
		XSendEvent(GDK_DISPLAY(),xwin,False,0,(XEvent *)&ev);
		break;
	    }
	}
	if (protocols != NULL) XFree(protocols);
    }
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(sgwin);
    }
}

void sgwindow_set_workspace_property(SGWindow *sgwin, guint32 val)
{
    Window xwin;

    if (sgwin == NULL) {
	xwin = GDK_ROOT_WINDOW();
    } else {
	GtkWidget *widget = GTK_WIDGET(sgwin);

	if (widget->window == NULL) return;
	xwin = GDK_WINDOW_XWINDOW(widget->window);
    }
    gdk_error_trap_push();
    XChangeProperty(GDK_DISPLAY(),xwin,xa_win_workspace,
      XA_CARDINAL,32,PropModeReplace,(unsigned char *)&val,1);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	;
    }
}

gint sgwindow_get_win_state(SGWindow *sgwin)
{
    gint rval = -1;
    Atom rt = None;
    int rf;
    unsigned long rni, bar;
    unsigned char *rv;
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return -1;
    gdk_error_trap_push();
    rt = 0;
    XGetWindowProperty(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(widget->window),
      xa_win_state,0,1,
      FALSE,XA_CARDINAL,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_CARDINAL) {
	if (rf == 32 && rni == 1) {
	    rval = ((gint *)rv)[0];
	}
	XFree(rv);
    }
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(sgwin);
    }
    return rval;
}

gint sgwindow_get_workspace_property(SGWindow *sgwin)
{
    gint rval = -1;
    Atom rt = None;
    int rf;
    unsigned long rni, bar;
    unsigned char *rv;
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return -1;
    gdk_error_trap_push();
    rt = 0;
    XGetWindowProperty(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(widget->window),
      xa_win_workspace,0,1,
      FALSE,XA_CARDINAL,&rt,&rf,&rni,&bar,&rv);
    if (rt == XA_CARDINAL) {
	if (rf == 32 && rni == 1) {
	    rval = ((gint *)rv)[0];
	}
	XFree(rv);
    }
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(sgwin);
    }
    return rval;
}

XWMHints *sgwindow_get_wm_hints(SGWindow *sgwin)
{
    XWMHints *p;
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return NULL;
    gdk_error_trap_push();
    p = XGetWMHints(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(widget->window));
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(sgwin);
	return NULL;
    }
    return p;
}

void sgwindow_get_geometry(SGWindow *sgwin, gint *x, gint *y, gint *w, gint *h,
  gint *bw)
{
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return;
    *x = widget->allocation.x;
    *y = widget->allocation.y;
    *w = widget->requisition.width;
    *h = widget->requisition.height;
    *bw = sgwin->border_width;
}

gboolean sgwindow_get_normal_hints(SGWindow *sgwin, XSizeHints *hints)
{
    long supplied;
    gboolean r;
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return FALSE;
    gdk_error_trap_push();
    r = XGetWMNormalHints(GDK_DISPLAY(),GDK_WINDOW_XWINDOW(widget->window),
      hints,&supplied) != 0;
    hints->flags &= supplied;
    if (gdk_error_trap_pop()) {
	r = FALSE;
    }
    return r;
}

void sgwindow_configure(SGWindow *sgwin, GtkAllocation *alloc,
  gint border_width, Window sibling, gint stack_mode, gint value_mask)
{
    XWindowChanges xwc;
    Window xwin;
    GtkWidget *widget = GTK_WIDGET(sgwin);

    if (widget->window == NULL) return;
    if (value_mask & CWX) {
	widget->allocation.x = alloc->x;
    }
    if (value_mask & CWY) {
	widget->allocation.y = alloc->y;
    }
    if (value_mask & CWWidth) {
	widget->requisition.width = widget->allocation.width = alloc->width;
    }
    if (value_mask & CWHeight) {
	widget->requisition.height = widget->allocation.height = alloc->height;
    }
    if (value_mask & CWBorderWidth) {
	sgwin->border_width = border_width;
    }
    xwin = GDK_WINDOW_XWINDOW(widget->window);
    gdk_error_trap_push();
    xwc.x = alloc->x;
    xwc.y = alloc->y;
    xwc.width = alloc->width;
    xwc.height = alloc->height;
    xwc.sibling = sibling;
    xwc.stack_mode = stack_mode;
    XConfigureWindow(GDK_DISPLAY(),xwin,value_mask,&xwc);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	return;
    }
}

void sgwindow_get_position(SGWindow *sgwin, gint *x, gint *y)
{
    GtkWidget *widget = GTK_WIDGET(sgwin);

    *x = widget->allocation.x;
    *y = widget->allocation.y;
}

gboolean sgwindow_ping(SGWindow *sgwin)
{
    GtkWidget *widget;
    Window rw,xid;
    gint x,y,w,h,bw;
    guint d;

    widget  = (GtkWidget *)sgwin;
    if (widget->window == NULL) return FALSE;
    xid = GDK_WINDOW_XWINDOW(widget->window);
    gdk_error_trap_push();
    XGetGeometry(GDK_DISPLAY(),xid,&rw,&x,&y,&w,&h,&bw,&d);
    gdk_flush();
    if (gdk_error_trap_pop()) {
	sgwindow_destroy_notify(SGWINDOW(widget));
	return FALSE;
    }
    return TRUE;
}
